/***********************************************
     Author :      Muddassar Farooq
     Organizaton:  The University of New South Wales

     Date Started : October 1, 1998
     Date Ended   : November 22 , 1998.
     Main Module : vncviewer.java

     other Modules : animatedMemoryImageSource.java,
     authenticationPanel.java,ButtonPanel.java,
     clipBoardFrame.java, DesCipher.java,optionsFrame.java,
     rfbProto.java, TimeEvent.java, TimeOut.java, vncCanvas.java,

     This programs has been modified under GNU license as allowed by                                      
     its distributor Olivetti & Oracle Research Laboratory. The purpose
     was to upgrade this software to new java 1.1 event model and get
     rid of all deprecated APIs. Moreover to make this software more an
     application than an applet. Following modifications are done in the applet
     
     1. One can start it as an application by typing
        java vncviewer host:display.
     2. One can send a middle click as a double click
     3. Scroll Pane is added to take care of large display area.
     4. Software is documented in a manner that conforms to principle
        of software Engineering so that its understanding and modifications
        become easy.


****************************************************************************/


//
// ButtonPannel.java : This is a panel that actually displays four buttons
// in vncviewer and implements action listener for them as required in
// new java event model 1.1, moreover this was required to actually make
// it a distinct part to achieve greater readibility/modularity.
//
//



import java.awt.*;
import java.awt.event.*;
import java.io.*;


public class ButtonPannel extends Panel implements ActionListener {


    Button disconnectButton;
    Button optionsButton;
    Button clipboardButton;
    Button ctrlAltDelButton;
    vncviewer app;


    public ButtonPannel(vncviewer v1) {
        app = v1;

        // Flow Layout is used for this panel

        setLayout(new FlowLayout(FlowLayout.LEFT, 0, 0));
        disconnectButton = new Button("Disconnect");

        // These commands actually add action listener for disconnect button and also
        // associate a special command word for that button.

        disconnectButton.addActionListener(this);
        disconnectButton.setActionCommand("Disconnect");

        disconnectButton.setEnabled(false);
        add(disconnectButton);


        optionsButton = new Button("Options");
        optionsButton.addActionListener(this);
        optionsButton.setActionCommand("Options");
        add(optionsButton);

        clipboardButton = new Button("Clipboard");
        clipboardButton.setEnabled(false);
        clipboardButton.addActionListener(this);
        clipboardButton.setActionCommand("Clipboard");
        add(clipboardButton);

        ctrlAltDelButton = new Button("Send Ctrl-Alt-Del");
        ctrlAltDelButton.setEnabled(false);
        ctrlAltDelButton.addActionListener(this);
        ctrlAltDelButton.setActionCommand("Ctr-Alt-Del");
        add(ctrlAltDelButton);

    }


    //
    // actionPerformed(ActionEvent e) : This method actually takes care of all
    // the event handling of associated with this panel. It simply finds out
    // which action command has been actually recieved and the does things
    // associated with them.

    public void actionPerformed(ActionEvent e) {

        String command = e.getActionCommand();

        // In this case simply show the options frame when clicked on this button.

        if (command.equals("Options")) {
            if (app.options.isVisible()) {
                app.options.setVisible(false);
            }
            else {
                app.options.setVisible(true);
            }
        }

        // Do all necessary things that needs to be done in order to do cleanup.
        // Some additional things are needed to be done once in an applet.

        else if (command.equals("Disconnect")) {

            System.out.println("disconnect");
            app.options.dispose();
            app.clipboard.dispose();

            if (app.inAnApplet) {
                app.removeAll();
                app.rfb.close();
                app.rfb = null;
                Label l = new Label("Disconnected");
                setLayout(new FlowLayout(FlowLayout.LEFT, 30, 30));
                add(l);
                validate();
                app.rfbThread.stop();
            }
            else {
                System.exit(1);
            }
        }

        // Displays the clipboard frame once cliked on this button.

        else if (command.equals("Clipboard")) {
            if (app.clipboard.isVisible()) {
                app.clipboard.setVisible(false);
            }
            else {
                app.clipboard.setVisible(true);
            }
        }


        // In case of click on this button simply create a keyevent up and down
        // and then send it. This event is created using the new event model.

        else if (command.equals("Ctr-Alt-Del")) {
            try {
                int modifiers = InputEvent.CTRL_MASK | InputEvent.ALT_MASK;

                KeyEvent altCtrDelDown =
                        new KeyEvent(this, KeyEvent.KEY_PRESSED, 0,
                        modifiers, 127);

                KeyEvent altCtrDelUp =
                        new KeyEvent(this, KeyEvent.KEY_RELEASED, 0,
                        modifiers, 127);

                app.rfb.writeKeyEvent(altCtrDelDown);
                app.rfb.writeKeyEvent(altCtrDelUp);

            } catch (IOException ioEx) {

                ioEx.printStackTrace();
            }
        }
    }
}