/***********************************************
     Author :      Muddassar Farooq
     Organizaton:  The University of New South Wales

     Date Started : October 1, 1998
     Date Ended   : November 22 , 1998.
     Main Module : vncviewer.java

     other Modules : animatedMemoryImageSource.java,
     authenticationPanel.java,ButtonPanel.java,
     clipBoardFrame.java, DesCipher.java,optionsFrame.java,
     rfbProto.java, TimeEvent.java, TimeOut.java, vncCanvas.java,

     This programs has been modified under GNU license as allowed by                                      
     its distributor Olivetti & Oracle Research Laboratory. The purpose
     was to upgrade this software to new java 1.1 event model and get
     rid of all deprecated APIs. Moreover to make this software more an
     application than an applet. Following modifications are done in the applet
     
     1. One can start it as an application by typing
        java vncviewer host:display.
     2. One can send a middle click as a double click
     3. Scroll Pane is added to take care of large display area.
     4. Software is documented in a manner that conforms to principle
        of software Engineering so that its understanding and modifications
        become easy.


****************************************************************************/

//
//  Copyright (C) 1997, 1998 Olivetti & Oracle Research Laboratory
//
//  This is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This software is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this software; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
//  USA.
//

//
//  animatedMemoryImageSource.java. This class implements an ImageProducer,
//  that is an interface for objects that can produce image data for images.Each
//  image contains an image producer which is needed to reconstruct the image
//  whenever it is needed with the help of createImage(). The image producer
//  provides data to an image consumer--an object that adheres to this
//  interface.
//


import java.awt.image.*;

class animatedMemoryImageSource implements ImageProducer {
    int width;
    int height;
    ColorModel cm;
    byte[] pixels;
    ImageConsumer ic;


    // animatedMemoryImageSource(int w,int h,ColorModel c, byte[] p). A simple
    // constructor for this class.

    animatedMemoryImageSource(int w, int h, ColorModel c, byte[] p) {
        width = w;
        height = h;
        cm = c;
        pixels = p;
    }


    //
    //  addConsumer(ImageConsumer c). This method is used to register an Image
    //  cosumer with the Image Producer for access to the image data during a
    //  later reconstruction.


    public void addConsumer(ImageConsumer c) {
        if (ic == c)
            return;

        if (ic != null) {

            //  when image producer is finsihed delivering all of pixels data that source
            //  image contains or a single frame.

            ic.imageComplete(ImageConsumer.IMAGEERROR);

        }

        ic = c;
        ic.setDimensions(width, height);
        ic.setColorModel(cm);
        ic.setHints(ImageConsumer.RANDOMPIXELORDER);  //   pixel data can be delivered
                                                                                                // any order.

        //      The pixels of image is delivered by making one or more calls
        //  to setPixels.
        ic.setPixels(0, 0, width, height, cm, pixels, 0, width);
        ic.imageComplete(ImageConsumer.SINGLEFRAMEDONE);
    }

    //
    //      isConsumer(ImageConsumer c) : This method determines if a given image
    //      consumer object is currently registered with this image producer.

    public boolean isConsumer(ImageConsumer c) {
        return (ic == c);
    }

    public void removeConsumer(ImageConsumer c) { //This method removes an image
                                                                                //      consumer.
        if (ic == c)
            ic = null;
    }

    public void requestTopDownLeftRightResend(ImageConsumer c) {
    }

    public void startProduction(ImageConsumer c) {
        addConsumer(c);
    }

    //  In setPixels an array of pixels contain image data. The formula for
    //  determining what entry in the pixels array contains data for
    //  pixel(x+i,y+j) is  offset+(j*scansize)+i where (0<=i<=w) and (0<=j<=h).



    void newPixels(int x, int y, int w, int h) {
        if (ic != null) {
            ic.setPixels(x, y, w, h, cm, pixels, width * y + x, width);
            ic.imageComplete(ImageConsumer.SINGLEFRAMEDONE);
        }
    }
}