/***********************************************
     Author :      Muddassar Farooq
     Organizaton:  The University of New South Wales

     Date Started : October 1, 1998
     Date Ended   : November 22 , 1998.
     Main Module : vncviewer.java

     other Modules : animatedMemoryImageSource.java,
     authenticationPanel.java,ButtonPanel.java,
     clipBoardFrame.java, DesCipher.java,optionsFrame.java,
     rfbProto.java, TimeEvent.java, TimeOut.java, vncCanvas.java,

     This programs has been modified under GNU license as allowed by                                      
     its distributor Olivetti & Oracle Research Laboratory. The purpose
     was to upgrade this software to new java 1.1 event model and get
     rid of all deprecated APIs. Moreover to make this software more an
     application than an applet. Following modifications are done in the applet
     
     1. One can start it as an application by typing
        java vncviewer host:display.
     2. One can send a middle click as a double click
     3. Scroll Pane is added to take care of large display area.
     4. Software is documented in a manner that conforms to principle
        of software Engineering so that its understanding and modifications
        become easy.


****************************************************************************/

//
//  Copyright (C) 1997, 1998 Olivetti & Oracle Research Laboratory
//
//  This is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This software is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this software; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
//  USA.
//

//
// Options frame.
//
// This deals with all the options the user can play with.
// It sets the encodings array and some booleans.
//

import java.awt.*;
import java.awt.event.*;

class optionsFrame extends Frame {

    vncviewer v;
    Label[] labels = new Label[names.length];
    Choice[] choices = new Choice[names.length];
    Button dismiss;

    static String[] names = { "Encoding", "Use CopyRect", "Raw pixel drawing",
    "CopyRect", "Share desktop", };

    static String[][] values = { { "Raw", "RRE", "CoRRE", "Hextile" }, { "Yes",
"No" }, { "Fast", "Reliable" }, { "Fast",
    "Reliable" }, { "Yes", "No" }, };

    final int encodingIndex = 0, useCopyRectIndex = 1,
    rawPixelDrawingIndex = 2, copyRectFastIndex = 3, shareDesktopIndex = 4;



    //
    // The actual data which other classes look at:
    //

    int[] encodings = new int[10];
    int nEncodings;


    boolean drawEachPixelForRawRects;

    boolean copyRectFast;

    boolean shareDesktop;


    //
    // Constructor.  Set up the labels and choices from the names and values
    // arrays.
    //

    optionsFrame(vncviewer v1) {
        super("VNC Options");

        v = v1;

        GridBagLayout gridbag = new GridBagLayout();
        setLayout(gridbag);

        GridBagConstraints gbc = new GridBagConstraints();
        gbc.fill = GridBagConstraints.BOTH;

        // add all top level labels in the frame.

        for (int i = 0; i < names.length; i++) {
            labels[i] = new Label(names[i]);
            gbc.gridwidth = 1;
            gridbag.setConstraints(labels[i], gbc);
            add(labels[i]);

            choices[i] = new Choice();
            gbc.gridwidth = GridBagConstraints.REMAINDER;
            gridbag.setConstraints(choices[i], gbc);
            add(choices[i]);

            //  for each of one of the actions a list of choices is added to the label.

            for (int j = 0; j < values[i].length; j++) {
                choices[i].addItem(values[i][j]);
            }
        }


        //      Define two listeners that will take care of all sort of actions.

        ItemListener setEncodingListener = new ItemListener() {
                    public void itemStateChanged(ItemEvent e) {
                        setEncodings();
                    }
                };

        ItemListener setOthersListener = new ItemListener() {
                    public void itemStateChanged(ItemEvent e) {
                        setOtherOptions();
                    }
                };

        //  Initially two options is being taken care of by first listener and later
        //  choices taken care of by second action listener.

        for (int i = 0; i < 2; i++) {
            choices[i].addItemListener(setEncodingListener);
        }

        for (int i = 2; i < choices.length; i++) {
            choices[i].addItemListener(setOthersListener);
        }

        dismiss = new Button("Dismiss");
        gbc.gridwidth = GridBagConstraints.REMAINDER;
        gridbag.setConstraints(dismiss, gbc);

        dismiss.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        setVisible(false);
                    }
                }
                );

        add(dismiss);

        pack();

        // Set up defaults

        choices[encodingIndex].select("Hextile");
        choices[useCopyRectIndex].select("Yes");
        choices[rawPixelDrawingIndex].select("Reliable");
        choices[copyRectFastIndex].select("Fast");
        choices[shareDesktopIndex].select("Yes");

        // But let them be overridden by parameters

        for (int i = 0; i < names.length; i++) {
            String s = v.readParameter(names[i], false);
            if (s != null) {
                for (int j = 0; j < values[i].length; j++) {
                    if (s.equalsIgnoreCase(values[i][j])) {
                        choices[i].select(j);
                    }
                }
            }
        }

        // Make the booleans and encodings array correspond to the state of the GUI

        setEncodings();
        setOtherOptions();
    }


    //
    // Disable shareDesktop option
    //

    void disableShareDesktop() {
        labels[shareDesktopIndex].setEnabled(false);
        choices[shareDesktopIndex].setEnabled(false);
    }

    //
    // setEncodings looks at the encoding and copyRect choices and sets the
    // encodings array appropriately.  It also calls the vncviewer's
    // setEncodings method to send a message to the RFB server if necessary.
    //

    void setEncodings() {
        nEncodings = 0;
        if (choices[useCopyRectIndex].getSelectedItem().equals("Yes")) {
            encodings[nEncodings++] = rfbProto.EncodingCopyRect;
        }

        int preferredEncoding = rfbProto.EncodingRaw;

        if (choices[encodingIndex].getSelectedItem().equals("RRE")) {
            preferredEncoding = rfbProto.EncodingRRE;
        } else if (choices[encodingIndex].getSelectedItem().equals("CoRRE")) {
            preferredEncoding = rfbProto.EncodingCoRRE;
        } else if (choices[encodingIndex].getSelectedItem().equals("Hextile")) {
            preferredEncoding = rfbProto.EncodingHextile;
        }

        if (preferredEncoding == rfbProto.EncodingRaw) {
            choices[rawPixelDrawingIndex].select("Fast");
            drawEachPixelForRawRects = false;
        }

        encodings[nEncodings++] = preferredEncoding;
        if (preferredEncoding != rfbProto.EncodingRRE) {
            encodings[nEncodings++] = rfbProto.EncodingRRE;
        }
        if (preferredEncoding != rfbProto.EncodingCoRRE) {
            encodings[nEncodings++] = rfbProto.EncodingCoRRE;
        }
        if (preferredEncoding != rfbProto.EncodingHextile) {
            encodings[nEncodings++] = rfbProto.EncodingHextile;
        }

        v.setEncodings();
    }

    //
    // setOtherOptions looks at the "other" choices (ones which don't set the
    // encoding) and sets the boolean flags appropriately.
    //

    void setOtherOptions() {

        drawEachPixelForRawRects = choices[rawPixelDrawingIndex]
                .getSelectedItem().equals("Reliable");

        copyRectFast =
                (choices[copyRectFastIndex].getSelectedItem().equals("Fast"));

        shareDesktop =
                (choices[shareDesktopIndex].getSelectedItem().equals("Yes"));
    }
}






