/***********************************************
     Author :      Muddassar Farooq
     Organizaton:  The University of New South Wales

     Date Started : October 1, 1998
     Date Ended   : November 22 , 1998.
     Main Module : vncviewer.java

     other Modules : animatedMemoryImageSource.java,
     authenticationPanel.java,ButtonPanel.java,
     clipBoardFrame.java, DesCipher.java,optionsFrame.java,
     rfbProto.java, TimeEvent.java, TimeOut.java, vncCanvas.java,

     This programs has been modified under GNU license as allowed by                                      
     its distributor Olivetti & Oracle Research Laboratory. The purpose
     was to upgrade this software to new java 1.1 event model and get
     rid of all deprecated APIs. Moreover to make this software more an
     application than an applet. Following modifications are done in the applet
     
     1. One can start it as an application by typing
        java vncviewer host:display.
     2. One can send a middle click as a double click
     3. Scroll Pane is added to take care of large display area.
     4. Software is documented in a manner that conforms to principle
        of software Engineering so that its understanding and modifications
        become easy.


****************************************************************************/


//
//  Copyright (C) 1997, 1998 Olivetti & Oracle Research Laboratory
//
//  This is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This software is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this software; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
//  USA.
//

//
// rfbProto.java.  This class actually models a complete vnc protcol. It
// defined all data variables that is needed for vnc protocol and it also
// defines all methods that actually implements vnc protocil.
//

import java.io.*;
import java.awt.*;
import java.awt.event.*;
import java.net.Socket;


class rfbProto {

    final String versionMsg = "RFB 003.003\n";
    final int ConnFailed = 0, NoAuth = 1, VncAuth = 2;
    final int VncAuthOK = 0, VncAuthFailed = 1, VncAuthTooMany = 2;

    final int FramebufferUpdate = 0, SetColourMapEntries = 1, Bell = 2,
    ServerCutText = 3;

    final int SetPixelFormat = 0, FixColourMapEntries = 1,
    SetEncodings = 2, FramebufferUpdateRequest = 3,
    KeyEvent = 4, PointerEvent = 5, ClientCutText = 6;

    final static int EncodingRaw = 0, EncodingCopyRect = 1,
    EncodingRRE = 2, EncodingCoRRE = 4, EncodingHextile = 5;

    final int HextileRaw = (1 << 0);
    final int HextileBackgroundSpecified = (1 << 1);
    final int HextileForegroundSpecified = (1 << 2);
    final int HextileAnySubrects = (1 << 3);
    final int HextileSubrectsColoured = (1 << 4);

    String host;
    int port;
    Socket sock;
    DataInputStream is;
    OutputStream os;
    boolean inNormalProtocol = false;
    vncviewer v;


    //
    // Constructor.  Just make TCP connection to RFB server.
    //

    rfbProto(String h, int p, vncviewer v1) throws IOException {
        v = v1;
        host = h;
        port = p;
        sock = new Socket(host, port);
        is = new DataInputStream(
                new BufferedInputStream(sock.getInputStream(), 16384));
        os = sock.getOutputStream();
    }


    void close() {
        try {
            sock.close();
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    //
    // Read server's protocol version message
    //

    int serverMajor, serverMinor;

    void readVersionMsg() throws IOException {

        byte[] b = new byte[12];

        is.readFully(b);

        if ((b[0] != 'R') || (b[1] != 'F') || (b[2] != 'B') ||
                (b[3] != ' ') || (b[4] < '0') || (b[4] > '9') ||
                (b[5] < '0') || (b[5] > '9') || (b[6] < '0') ||
                (b[6] > '9') || (b[7] != '.') || (b[8] < '0') ||
                (b[8] > '9') || (b[9] < '0') || (b[9] > '9') ||
                (b[10] < '0') || (b[10] > '9') || (b[11] != '\n')) {
            throw new IOException("Host " + host + " port " + port + " is not an RFB server");
        }

        serverMajor = (b[4] - '0') * 100 + (b[5] - '0') * 10 + (b[6] - '0');
        serverMinor =  (b[8] - '0') * 100 + (b[9] - '0') * 10 + (b[10] - '0');

    }


    //
    // Write our protocol version message
    //

    void writeVersionMsg() throws IOException {
        byte[] b = new byte[12];
        b = versionMsg.getBytes();
        os.write(b);
    }


    //
    // Find out the authentication scheme.
    //

    int readAuthScheme() throws IOException {
        int authScheme = is.readInt();

        switch (authScheme) {

            case ConnFailed:
                int reasonLen = is.readInt();
                byte[] reason = new byte[reasonLen];
                is.readFully(reason);

                throw new IOException(new String(reason));

            case NoAuth:
            case VncAuth:
                return authScheme;

            default:
                throw new IOException(
                        "Unknown authentication scheme from RFB " +    "server " + authScheme);

        }
    }


    //
    // Write the client initialisation message
    //

    void writeClientInit() throws IOException {
        if (v.options.shareDesktop) {
            os.write(1);

        } else {
            os.write(0);

        }
        v.options.disableShareDesktop();
    }


    //
    // Read the server initialisation message
    //

    String desktopName;
    int framebufferWidth, framebufferHeight;
    int bitsPerPixel, depth;
    boolean bigEndian, trueColour;
    int redMax, greenMax, blueMax, redShift, greenShift, blueShift;

    void readServerInit() throws IOException {


        framebufferWidth = is.readUnsignedShort();
        framebufferHeight = is.readUnsignedShort();

        bitsPerPixel = is.readUnsignedByte();
        depth = is.readUnsignedByte();
        bigEndian = (is.readUnsignedByte() != 0);
        trueColour = (is.readUnsignedByte() != 0);


        redMax = is.readUnsignedShort();
        greenMax = is.readUnsignedShort();
        blueMax = is.readUnsignedShort();

        redShift = is.readUnsignedByte();
        greenShift = is.readUnsignedByte();
        blueShift = is.readUnsignedByte();

        byte[] pad = new byte[3];
        is.read(pad);
        int nameLength = is.readInt();

        byte[] name = new byte[nameLength];
        is.readFully(name);
        desktopName = new String(name);

        inNormalProtocol = true;
    }


    //
    // Read the server message type
    //

    int readServerMessageType() throws IOException {
        return is.read();
    }


    //
    // Read a FramebufferUpdate message
    //

    int updateNRects;

    void readFramebufferUpdate() throws IOException {
        is.readByte();
        updateNRects = is.readUnsignedShort();
    }

    // Read a FramebufferUpdate rectangle header

    int updateRectX, updateRectY, updateRectW, updateRectH,
    updateRectEncoding;

    void readFramebufferUpdateRectHdr() throws IOException {
        updateRectX = is.readUnsignedShort();
        updateRectY = is.readUnsignedShort();

        updateRectW = is.readUnsignedShort();
        updateRectH = is.readUnsignedShort();

        updateRectEncoding = is.readInt();

        if ((updateRectX + updateRectW > framebufferWidth) ||
                (updateRectY + updateRectH > framebufferHeight)) {
            throw new IOException(
                    "Framebuffer update rectangle too large: " +
                    updateRectW + "x" + updateRectH + " at (" +
                    updateRectX + "," + updateRectY + ")");
        }
    }

    // Read CopyRect source X and Y.

    int copyRectSrcX, copyRectSrcY;

    void readCopyRect() throws IOException {
        copyRectSrcX = is.readUnsignedShort();
        copyRectSrcY = is.readUnsignedShort();

    }


    //
    // Read a ServerCutText message
    //

    String readServerCutText() throws IOException {
        byte[] pad = new byte[3];
        is.read(pad);
        int len = is.readInt();
        byte[] text = new byte[len];
        is.readFully(text);
        return new String(text);
    }


    //
    // Write a FramebufferUpdateRequest message
    //

    void writeFramebufferUpdateRequest(int x, int y, int w, int h,
            boolean incremental) throws IOException {
        byte[] b = new byte[10];

        b[0] = (byte) FramebufferUpdateRequest;
        b[1] = (byte)(incremental ? 1 : 0);
        b[2] = (byte)((x >> 8) & 0xff);
        b[3] = (byte)(x & 0xff);
        b[4] = (byte)((y >> 8) & 0xff);
        b[5] = (byte)(y & 0xff);
        b[6] = (byte)((w >> 8) & 0xff);
        b[7] = (byte)(w & 0xff);
        b[8] = (byte)((h >> 8) & 0xff);
        b[9] = (byte)(h & 0xff);

        os.write(b);
    }


    //
    // Write a SetPixelFormat message
    //

    void writeSetPixelFormat(int bitsPerPixel, int depth, boolean bigEndian, boolean trueColour, int redMax,
                                           int greenMax, int blueMax, int redShift, int greenShift,
                                           int blueShift) throws IOException {
        byte[] b = new byte[20];

        b[0] = (byte) SetPixelFormat;
        b[4] = (byte) bitsPerPixel;
        b[5] = (byte) depth;
        b[6] = (byte)(bigEndian ? 1 : 0);
        b[7] = (byte)(trueColour ? 1 : 0);
        b[8] = (byte)((redMax >> 8) & 0xff);
        b[9] = (byte)(redMax & 0xff);
        b[10] = (byte)((greenMax >> 8) & 0xff);
        b[11] = (byte)(greenMax & 0xff);
        b[12] = (byte)((blueMax >> 8) & 0xff);
        b[13] = (byte)(blueMax & 0xff);
        b[14] = (byte) redShift;
        b[15] = (byte) greenShift;
        b[16] = (byte) blueShift;

        os.write(b);

    }


    //
    // Write a FixColourMapEntries message.  The values in the red, green and
    // blue arrays are from 0 to 65535.
    //

    void writeFixColourMapEntries(int firstColour, int nColours, int[] red, int[] green, int[] blue)
      throws IOException {
        byte[] b = new byte[6 + nColours * 6];

        b[0] = (byte) FixColourMapEntries;
        b[2] = (byte)((firstColour >> 8) & 0xff);
        b[3] = (byte)(firstColour & 0xff);
        b[4] = (byte)((nColours >> 8) & 0xff);
        b[5] = (byte)(nColours & 0xff);

        for (int i = 0; i < nColours; i++) {
            b[6 + i * 6] = (byte)((red[i] >> 8) & 0xff);
            b[6 + i * 6 + 1] = (byte)(red[i] & 0xff);
            b[6 + i * 6 + 2] = (byte)((green[i] >> 8) & 0xff);
            b[6 + i * 6 + 3] = (byte)(green[i] & 0xff);
            b[6 + i * 6 + 4] = (byte)((blue[i] >> 8) & 0xff);
            b[6 + i * 6 + 5] = (byte)(blue[i] & 0xff);
        }

        os.write(b);
    }


    //
    // Write a SetEncodings message
    //

    void writeSetEncodings(int[] encs, int len) throws IOException {
        byte[] b = new byte[4 + 4 * len];

        b[0] = (byte) SetEncodings;
        b[2] = (byte)((len >> 8) & 0xff);
        b[3] = (byte)(len & 0xff);

        for (int i = 0; i < len; i++) {
            b[4 + 4 * i] = (byte)((encs[i] >> 24) & 0xff);
            b[5 + 4 * i] = (byte)((encs[i] >> 16) & 0xff);
            b[6 + 4 * i] = (byte)((encs[i] >> 8) & 0xff);
            b[7 + 4 * i] = (byte)(encs[i] & 0xff);
        }

        os.write(b);
    }


    //
    // Write a ClientCutText message
    //

    void writeClientCutText(String text) throws IOException {
        int offset = 8;
        int length = text.length();

        byte[] tempb = new byte[length];
        byte[] b = new byte[offset + length];


        b[0] = (byte) ClientCutText;
        b[4] = (byte)((text.length() >> 24) & 0xff);
        b[5] = (byte)((text.length() >> 16) & 0xff);
        b[6] = (byte)((text.length() >> 8) & 0xff);
        b[7] = (byte)(text.length() & 0xff);

        tempb = text.getBytes();
        for (int i = 0; i < length; i++) {
            b[offset + i] = tempb[i];
        }

        os.write(b);
    }


    //
    // A buffer for putting pointer and keyboard Events before being sent.  This
    // is to ensure that multiple RFB Events generated from a single Java Event
    // will all be sent in a single network packet.  The maximum possible
    // length is 4 modifier down InputEvents, a single key InputEvent followed by 4
    // modifier up InputEvents i.e. 9 key InputEvents or 72 bytes.
    //

    byte[] eventBuf = new byte[72];
    int eventBufLen;


    //
    // Write a pointer InputEvent message.  We may need to send modifier key InputEvents
    // around it to set the correct modifier state.  Also buttons 2 and 3 are
    // represented as having ALT and META modifiers respectively. Moreover
    // middleid specifies that whether an actually middle button is pressed on
    // 3 mouse button. In case of an actuall middle click we need not to
    // ignore any click following an up event.
    //

    int pointerMask = 0;


    void writePointerEvent(AWTEvent e,
            boolean middleid) throws IOException {

        MouseEvent localme = (MouseEvent) e;
        int evtmodifiers = localme.getModifiers();
        int x = localme.getX();
        int y = localme.getY();


        byte[] b = new byte[6];

        if (localme.getID() == MouseEvent.MOUSE_PRESSED) {
            if (middleid)
                pointerMask = 2; // This is the mask for middle button

            else {
                if (localme.isMetaDown())
                    pointerMask = 4;
                else
                    pointerMask = 1;
            }

        } else if (localme.getID() == MouseEvent.MOUSE_RELEASED) {
            pointerMask = 0;
        }

        evtmodifiers &= ~(InputEvent.ALT_MASK | InputEvent.META_MASK);

        eventBufLen = 0;

        writeModifierKeyEvents(evtmodifiers);

        if (x < 0)
            x = 0;
        if (y < 0)
            y = 0;

        eventBuf[eventBufLen++] = (byte) PointerEvent;
        eventBuf[eventBufLen++] = (byte) pointerMask;
        eventBuf[eventBufLen++] = (byte)((x >> 8) & 0xff);
        eventBuf[eventBufLen++] = (byte)(x & 0xff);
        eventBuf[eventBufLen++] = (byte)((y >> 8) & 0xff);
        eventBuf[eventBufLen++] = (byte)(y & 0xff);

        //
        // Always release all modifiers after an "up" event
        //

        if (pointerMask == 0) {
            writeModifierKeyEvents(0);
        }

        os.write(eventBuf, 0, eventBufLen);
    }



    //
    // Write a key Event message.  We may need to send modifier key Events
    // around it to set the correct modifier state.  Also we need to translate
    // from the Java key values to the X keysym values used by the RFB protocol.
    //

    void writeKeyEvent(AWTEvent e) throws IOException {

        KeyEvent localkey = (KeyEvent) e;
        if (isModDown(localkey))  //  We need not to send a modifier key because
            return;                           //  it will be sent as writemodifierkeys.


        int key;
        int evtmodifiers = localkey.getModifiers();
        boolean down = false;

        if (localkey.getID() == localkey.KEY_PRESSED)
            down = true;

        if (localkey.isActionKey()) { //if an action key then simply find its code
            key = localkey.getKeyCode();

            //
            // A KEY_ACTION Event should be one of the following.  If not then just
            // ignore the Event.
            //

            switch (key) {
                case localkey.VK_HOME:
                    key = 0xff50;
                    break;
                case localkey.VK_LEFT:
                    key = 0xff51;
                    break;
                case localkey.VK_UP:
                    key = 0xff52;
                    break;
                case localkey.VK_RIGHT:
                    key = 0xff53;
                    break;
                case localkey.VK_DOWN:
                    key = 0xff54;
                    break;
                case localkey.VK_PAGE_DOWN:
                    key = 0xff55;
                    break;
                case localkey.VK_PAGE_UP:
                    key = 0xff56;
                    break;
                case localkey.VK_END:
                    key = 0xff57;
                    break;
                case localkey.VK_F1:
                    key = 0xffbe;
                    break;
                case localkey.VK_F2:
                    key = 0xffbf;
                    break;
                case localkey.VK_F3:
                    key = 0xffc0;
                    break;
                case localkey.VK_F4:
                    key = 0xffc1;
                    break;
                case localkey.VK_F5:
                    key = 0xffc2;
                    break;
                case localkey.VK_F6:
                    key = 0xffc3;
                    break;
                case localkey.VK_F7:
                    key = 0xffc4;
                    break;
                case localkey.VK_F8:
                    key = 0xffc5;
                    break;
                case localkey.VK_F9:
                    key = 0xffc6;
                    break;
                case localkey.VK_F10:
                    key = 0xffc7;
                    break;
                case localkey.VK_F11:
                    key = 0xffc8;
                    break;
                case localkey.VK_F12:
                    key = 0xffc9;
                    break;

                default:
                    return;
            }

        } else {

            char iskey = localkey.getKeyChar(); //  If it is not an action key
            //  then determine its code.

            //
            // A "normal" key press.  Ordinary ASCII characters go straight through.
            // For CTRL-<letter>, CTRL is sent separately so just send <letter>.
            // Backspace, tab, return, escape and delete have special keysyms.
            // Anything else we ignore.
            //

            key = (int) iskey;

            if (key < 32) {
                if (localkey.isControlDown()) {
                    key += 96;
                } else {

                    switch (key) {
                        case localkey.VK_BACK_SPACE:
                            key = 0xff08;
                            break;
                        case localkey.VK_TAB:
                            key = 0xff09;
                            break;
                        case localkey.VK_ENTER:
                            key = 0xff0d;
                            break;
                        case localkey.VK_ESCAPE:
                            key = 0xff1b;
                            break;
                    }
                }

            } else if (key >= 127) {

                if (key == 127) {
                    key = 0xffff;

                } else {

                    // JDK1.1 on X incorrectly passes some keysyms straight through, so
                    // we do too.  JDK1.1.4 seems to have fixed this.
                    if ((key < 0xff00) || (key > 0xffff))
                        return;
                }
            }
        }

        eventBufLen = 0;
        writeModifierKeyEvents(evtmodifiers); //write modifier events.
        writeKeyEvent(key, down);

        //
        // Always release all modifiers after an "up" Event
        //

        if (!down) {
            writeModifierKeyEvents(0);
        }

        os.write(eventBuf, 0, eventBufLen);
    }


    public boolean isModDown(KeyEvent temp) {

        char code = temp.getKeyChar();

        if (code == temp.CHAR_UNDEFINED) {
            int ncode = temp.getKeyCode();
            if ((ncode == temp.VK_ALT) || (ncode == temp.VK_CONTROL) ||
                    (ncode == temp.VK_META) || (ncode == temp.VK_SHIFT))

                return true;

            else
                return false;
        }
        else
            return false;
    }


    //
    // Add a raw key Event with the given X keysym to EventBuf.
    //

    void writeKeyEvent(int keysym, boolean down) throws IOException {
        eventBuf[eventBufLen++] = (byte) KeyEvent;
        eventBuf[eventBufLen++] = (byte)(down ? 1 : 0);
        eventBuf[eventBufLen++] = (byte) 0;
        eventBuf[eventBufLen++] = (byte) 0;
        eventBuf[eventBufLen++] = (byte)((keysym >> 24) & 0xff);
        eventBuf[eventBufLen++] = (byte)((keysym >> 16) & 0xff);
        eventBuf[eventBufLen++] = (byte)((keysym >> 8) & 0xff);
        eventBuf[eventBufLen++] = (byte)(keysym & 0xff);
    }


    //
    // Write key Events to set the correct modifier state. In this case if
    // previously a key was pressed then it is not sent. Only condition to
    // write a key event is to find out that a new modifier is not same
    // as previous one.
    //

    int oldModifiers;

    void writeModifierKeyEvents(int newModifiers) throws IOException {
        if ((newModifiers & InputEvent.CTRL_MASK) !=
                (oldModifiers & InputEvent.CTRL_MASK))
            writeKeyEvent(0xffe3, (newModifiers & InputEvent.CTRL_MASK) != 0);

        if ((newModifiers & InputEvent.SHIFT_MASK) !=
                (oldModifiers & InputEvent.SHIFT_MASK))
            writeKeyEvent(0xffe1, (newModifiers & InputEvent.SHIFT_MASK) != 0);

        if ((newModifiers & InputEvent.META_MASK) !=
                (oldModifiers & InputEvent.META_MASK))
            writeKeyEvent(0xffe7, (newModifiers & InputEvent.META_MASK) != 0);

        if ((newModifiers & InputEvent.ALT_MASK) !=
                (oldModifiers & InputEvent.ALT_MASK))
            writeKeyEvent(0xffe9,(newModifiers & InputEvent.ALT_MASK) != 0);

        oldModifiers = newModifiers;
    }
}