/***********************************************
     Author :      Muddassar Farooq
     Organizaton:  The University of New South Wales

     Date Started : October 1, 1998
     Date Ended   : November 22 , 1998.
     Main Module : vncviewer.java

     other Modules : animatedMemoryImageSource.java,
     authenticationPanel.java,ButtonPanel.java,
     clipBoardFrame.java, DesCipher.java,optionsFrame.java,
     rfbProto.java, TimeEvent.java, TimeOut.java, vncCanvas.java,

     This programs has been modified under GNU license as allowed by                                      
     its distributor Olivetti & Oracle Research Laboratory. The purpose
     was to upgrade this software to new java 1.1 event model and get
     rid of all deprecated APIs. Moreover to make this software more an
     application than an applet. Following modifications are done in the applet
     
     1. One can start it as an application by typing
        java vncviewer host:display.
     2. One can send a middle click as a double click
     3. Scroll Pane is added to take care of large display area.
     4. Software is documented in a manner that conforms to principle
        of software Engineering so that its understanding and modifications
        become easy.


****************************************************************************/

//
//  Copyright (C) 1997, 1998 Olivetti & Oracle Research Laboratory
//
//  This is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This software is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this software; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
//  USA.
//

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.io.*;


//
//      vncCanvas is a subclass of Canvas actually takes care of drawing all
//      remote desktop data in a canvas. Moreover it acually also takes care of
//      all the event handling which has been occuring in the canvas. It actually
//      recieves those events and send them to remot vnc server. All event
//      handling has been upgraded to new 1.1 model.
//

class vncCanvas extends Canvas  // implements KeyListener
{
  vncviewer v;
  rfbProto rfb;
  animatedMemoryImageSource amis;
  TimeEvent canvaste;

  Color[] colors;
  ColorModel cm;

  Image rawPixelsImage;
  Image paintImage;

  byte[] pixels;
  boolean needToResetClip;

  Graphics sg, sg2;
  Graphics pig, pig2;

  vncCanvas(vncviewer v1) throws IOException {
    v = v1;
    rfb = v.rfb;

    //      In Java 1.1 event model all events to which a component listens are
    //      to be enabled by calling this method. Here RESERVED_ID_MAX + 2 serves
    //      as the time out event.

    this.enableEvents(AWTEvent.MOUSE_EVENT_MASK|
		      AWTEvent.MOUSE_MOTION_EVENT_MASK |
		      AWTEvent.KEY_EVENT_MASK|
		      AWTEvent.RESERVED_ID_MAX+2);

    this.requestFocus();
    cm = new DirectColorModel(8, 7, (7 << 3), (3 << 6));

    canvaste = new TimeEvent(this,AWTEvent.RESERVED_ID_MAX +2);

    rfb.writeSetPixelFormat(8, 8, false, true, 7, 7, 3, 0, 3, 6);

    colors = new Color[256];

    for (int i = 0; i < 256; i++) {
      colors[i] = new Color(cm.getRGB(i));
    }

    pixels = new byte[rfb.framebufferWidth * rfb.framebufferHeight];

    amis = new animatedMemoryImageSource(rfb.framebufferWidth,
					 rfb.framebufferHeight, cm, pixels);
    rawPixelsImage = createImage(amis);

    paintImage = v.createImage(rfb.framebufferWidth, rfb.framebufferHeight);

    pig = paintImage.getGraphics();
  }

  //      Following two methods are actually used in 1.1 event model.

  public Dimension getPreferredSize() {
    return new Dimension(rfb.framebufferWidth, rfb.framebufferHeight);
  }

  public Dimension getMinimumSize() {
    return new Dimension(rfb.framebufferWidth, rfb.framebufferHeight);
  }

  public void update(Graphics g) {
  }

  public void paint(Graphics g) {
    requestFocus();
    g.drawImage(paintImage,0, 0, this);
  }

  //
  // processNormalProtocol() - executed by the rfbThread to deal with the
  // RFB socket. This method actually reads all the required data and
  // simplay paints image on canvas, depending upon the encoding that has
  // been selected.
  //

  public void processNormalProtocol() throws IOException {

    rfb.writeFramebufferUpdateRequest(0, 0, rfb.framebufferWidth,
				      rfb.framebufferHeight, false);

    sg = getGraphics();

    needToResetClip = false;

    //
    // main dispatch loop
    //

    while (true) {
      int msgType = rfb.readServerMessageType();

      switch (msgType) {
      case rfb.FramebufferUpdate:
	rfb.readFramebufferUpdate();

	for (int i = 0; i < rfb.updateNRects; i++) {
	  rfb.readFramebufferUpdateRectHdr();

	  if (needToResetClip && (rfb.updateRectEncoding != rfb.EncodingRaw)) {
	    try {
	      sg.setClip(0, 0, rfb.framebufferWidth, rfb.framebufferHeight);
	      pig.setClip(0, 0, rfb.framebufferWidth, rfb.framebufferHeight);
	    } catch (NoSuchMethodError e) {
	    }
	    needToResetClip = false;
	  }

	  switch (rfb.updateRectEncoding) {

	  case rfb.EncodingRaw:
	    drawRawRect(rfb.updateRectX, rfb.updateRectY,
			rfb.updateRectW, rfb.updateRectH);
	    break;

	  case rfb.EncodingCopyRect:
	    rfb.readCopyRect();
	    pig.copyArea(rfb.copyRectSrcX, rfb.copyRectSrcY,
			 rfb.updateRectW, rfb.updateRectH,
			 rfb.updateRectX - rfb.copyRectSrcX,
			 rfb.updateRectY - rfb.copyRectSrcY);
	    if (v.options.copyRectFast) {
	      sg.copyArea(rfb.copyRectSrcX, rfb.copyRectSrcY,
			  rfb.updateRectW, rfb.updateRectH,
			  rfb.updateRectX - rfb.copyRectSrcX,
			  rfb.updateRectY - rfb.copyRectSrcY);
	    } else {
	      sg.drawImage(paintImage, 0, 0, this);
	    }
	    break;

	  case rfb.EncodingRRE:
	    {
	      int nSubrects = rfb.is.readInt();
	      int bg = rfb.is.read();


	      int pixel, x, y, w, h;
	      sg.translate(rfb.updateRectX,rfb.updateRectY);
	      sg.setColor(colors[bg]);
	      sg.fillRect(0, 0, rfb.updateRectW,rfb.updateRectH);
	      pig.translate(rfb.updateRectX,rfb.updateRectY);
	      pig.setColor(colors[bg]);
	      pig.fillRect(0, 0, rfb.updateRectW,rfb.updateRectH);

	      for (int j = 0; j < nSubrects; j++) {
		pixel = rfb.is.read();
		x = rfb.is.readUnsignedShort();
		y = rfb.is.readUnsignedShort();
		w = rfb.is.readUnsignedShort();
		h = rfb.is.readUnsignedShort();

		sg.setColor(colors[pixel]);
		sg.fillRect(x, y, w, h);
		pig.setColor(colors[pixel]);
		pig.fillRect(x, y, w, h);
	      }

	      sg.translate(-rfb.updateRectX,-rfb.updateRectY);
	      pig.translate(-rfb.updateRectX,-rfb.updateRectY);
	      break;
	    }

	  case rfb.EncodingCoRRE:
	    {
	      int nSubrects = rfb.is.readInt();
	      int bg = rfb.is.read();
	      int pixel, x, y, w, h;

	      sg.translate(rfb.updateRectX,rfb.updateRectY);
	      sg.setColor(colors[bg]);
	      sg.fillRect(0, 0, rfb.updateRectW,rfb.updateRectH);
	      pig.translate(rfb.updateRectX,rfb.updateRectY);
	      pig.setColor(colors[bg]);
	      pig.fillRect(0, 0, rfb.updateRectW,rfb.updateRectH);

	      for (int j = 0; j < nSubrects; j++) {
		pixel = rfb.is.read();
		x = rfb.is.read();
		y = rfb.is.read();
		w = rfb.is.read();
		h = rfb.is.read();
		sg.setColor(colors[pixel]);
		sg.fillRect(x, y, w, h);
		pig.setColor(colors[pixel]);
		pig.fillRect(x, y, w, h);
	      }

	      sg.translate(-rfb.updateRectX,-rfb.updateRectY);
	      pig.translate(-rfb.updateRectX,-rfb.updateRectY);

	      break;
	    }


	  case rfb.EncodingHextile:
	    {
	      int bg = 0, fg = 0, sx, sy, sw, sh;

	      for (int ty = rfb.updateRectY;
		   ty < rfb.updateRectY + rfb.updateRectH;
		   ty += 16) {

		for (int tx = rfb.updateRectX;
		     tx < rfb.updateRectX +rfb.updateRectW;
		     tx += 16) {

		  int tw = 16, th = 16;

		  if (rfb.updateRectX +rfb.updateRectW - tx < 16)
		    tw = rfb.updateRectX +rfb.updateRectW - tx;

		  if (rfb.updateRectY +rfb.updateRectH - ty < 16)
		    th = rfb.updateRectY +rfb.updateRectH - ty;

		  int subencoding = rfb.is.read();

		  if ((subencoding & rfb.HextileRaw) != 0) {
		    drawRawRect(tx, ty, tw, th);
		    continue;
		  }

		  if (needToResetClip) {
		    try {
		      sg.setClip(0,0,rfb.framebufferWidth, rfb.framebufferHeight);
		      pig.setClip(0,0,rfb.framebufferWidth, rfb.framebufferHeight);
		    } catch (NoSuchMethodError e) {
		    }
		    needToResetClip = false;
		  }


		  if ((subencoding & rfb.HextileBackgroundSpecified) != 0)
		    bg = rfb.is.read();

		  sg.setColor(colors[bg]);
		  sg.fillRect(tx, ty, tw, th);
		  pig.setColor(colors[bg]);
		  pig.fillRect(tx, ty, tw, th);

		  if ((subencoding & rfb.HextileForegroundSpecified) != 0)
		    fg = rfb.is.read();

		  if ((subencoding & rfb.HextileAnySubrects) == 0)
		    continue;

		  int nSubrects = rfb.is.read();

		  sg.translate(tx, ty);
		  pig.translate(tx, ty);

		  if ((subencoding & rfb.HextileSubrectsColoured) != 0) {

		    for (int j = 0; j < nSubrects; j++) {
		      fg = rfb.is.read();
		      int b1 = rfb.is.read();
		      int b2 = rfb.is.read();
		      sx = b1 >> 4;
		      sy = b1 & 0xf;
		      sw = (b2 >> 4) + 1;
		      sh = (b2 & 0xf) + 1;

		      sg.setColor(colors[fg]);
		      sg.fillRect(sx, sy, sw, sh);
		      pig.setColor(colors[fg]);
		      pig.fillRect(sx, sy, sw,sh);
		    }

		  } else {

		    sg.setColor(colors[fg]);
		    pig.setColor(colors[fg]);

		    for (int j = 0; j < nSubrects; j++) {
		      int b1 = rfb.is.read();
		      int b2 = rfb.is.read();
		      sx = b1 >> 4;
		      sy = b1 & 0xf;
		      sw = (b2 >> 4) + 1;
		      sh = (b2 & 0xf) + 1;

		      sg.fillRect(sx, sy, sw, sh);
		      pig.fillRect(sx, sy, sw,sh);
		    }
		  }


		  sg.translate(-tx, -ty);
		  pig.translate(-tx, -ty);
		}
	      }
	      break;
	    }

	  default:
	    throw new IOException("Unknown RFB rectangle encoding " + rfb.updateRectEncoding);
	  } // end switch
	}// end for

	rfb.writeFramebufferUpdateRequest(0, 0,rfb.framebufferWidth, rfb.framebufferHeight,true);
	break;

      case rfb.SetColourMapEntries:
	throw new IOException("Can't handle SetColourMapEntries message");

      case rfb.Bell:
	System.out.print((char)7);
	break;

      case rfb.ServerCutText:
	String s = rfb.readServerCutText();
	v.clipboard.setCutText(s);
	break;

      default:
	throw new IOException("Unknown RFB message type " + msgType);
      }
    }
  }


  //
  //    Draw a raw rectangle. This method has been used whe raw encoding type
  //    has     been chosen.
  //

  void drawRawRect(int x, int y, int w, int h) throws IOException {
    if (v.options.drawEachPixelForRawRects) {
      for (int j = y; j < (y + h); j++) {
	for (int k = x; k < (x + w); k++) {
	  int pixel = rfb.is.read();
	  sg.setColor(colors[pixel]);
	  sg.fillRect(k, j, 1, 1);
	  pig.setColor(colors[pixel]);
	  pig.fillRect(k, j, 1, 1);
	}
      }
      return;
    }

    for (int j = y; j < (y + h); j++) {
      rfb.is.readFully(pixels, j * rfb.framebufferWidth + x, w);
    }

    amis.newPixels(x, y, w, h);

    try {
      sg.setClip(x, y, w, h);
      pig.setClip(x, y, w, h);
      needToResetClip = true;
    } catch (NoSuchMethodError e) {
      sg2 = sg.create();
      sg.clipRect(x, y, w, h);
      pig2 = pig.create();
      pig.clipRect(x, y, w, h);
    }

    sg.drawImage(rawPixelsImage, 0, 0, this);
    pig.drawImage(rawPixelsImage, 0, 0, this);

    if (sg2 != null) {
      sg.dispose();    // reclaims resources more quickly
      sg = sg2;
      sg2 = null;
      pig.dispose();
      pig = pig2;
      pig2 = null;
    }
  }



  //
  // processEvent(AWTEvent e)
  //
  // Because of a "feature" in the AWT implementation over X, the vncCanvas
  // sometimes loses focus and the only way to get it back is to call
  // requestFocus() explicitly.  However we need to be careful when calling
  // requestFocus() on Windows or other click-to-type systems.  What we do is
  // call requestFocus() whenever there is mouse movement over the window,
  // AND the focus is already in the applet.


  //    Actually major upgrade in this has been to send a double click as a
  //    middle click and this issue has been complicated because of a bug
  //    till jdk1.1.6 i.e in which BUTTON1_MASK does not actually set when
  //    mouse is pressed on Unix, however the mask is accurately set
  //    in Windows when getModifiers is called. Because of this bug, an
  //    alternative way of actually doing has been evolved and that has
  //    made the algorithm a bit complex and lengthy.

  //          Algorithm is very simple that whole process undergoes three states
  //    listenstate, checkleftclick and checkrightclick. Initially it is
  //    set to listen state in which it listens for a click. The moment a mouse
  //    is cliked , depending upon the type state is changed. If it is middle
  //    click then state goes back to listen.

  //    The trick was that once a rightbutton is pressed follwed by left
  //    button then that cannot be detected through isAlDown() as
  //    in our case defualt case is for left click. In order to avoid
  //    this thing a falg previousMetaPressed has been defined. If meta is set
  //    now and meta was also previouly set without actually a mouse release
  //    event then that has been acknowledged as a left click. This is
  //    the logic which has been complex.

  //    Finally if left/right or roght/left click do not occur within
  //    timeout time then they are interpreted as two different clicks. and
  //    state is set to listen state.

  //    In case of a middle clik we have to send a middle click press followed
  //    by middleup event. That means after double click we have to ignore
  //    one up event.


  final int listenstate=1,checkleftclick=2,checkrightclick=3;// different states
  int processstate=listenstate;

  boolean processingPressMouse = false;
  boolean previousMetaPressed = false; //  A flag that helps in finding double
  //     click
  boolean middleclick = false;
  boolean ignorenextclick = false;     // A flag that helps us in sending middle
  //      up events.

  MouseEvent firstClick,secondClick;

  Math test; // For using abs functions.
  TimeOut theTimeOut;                   // TimeOut Event defined.


  public void  processEvent(AWTEvent e) {
    if ((rfb != null) && rfb.inNormalProtocol) {
      try {
	switch (e.getID()) {

	case MouseEvent.MOUSE_PRESSED  :
	  processingPressMouse = true;
	  MouseEvent me = (MouseEvent) e;

	  // if focus lost then request it.

	  if (v.gotFocus) {
	    requestFocus();
	  }

	  switch(processstate) {

	  case listenstate :

                                // if middle button click then send it.

	    if(me.isAltDown())  {

	      firstClick =  me;
	      rfb.writePointerEvent(firstClick,true);
	      processstate = listenstate;

	      // if right button click then look for left click .

	    } else if( me.isMetaDown()) {
	      firstClick =  me;
	      processstate = checkleftclick;
	      previousMetaPressed = true;
	      // if left button click then look for right click .

	    } else {
	      firstClick =  me;
	      processstate = checkrightclick;
	    }


                                // Set a timeout event of 30 msec, and start it.

	    theTimeOut= new TimeOut(30,this);
	    theTimeOut.start();
	    break;

	  case checkleftclick :

                                //  if looking for left click and middle button pressed
                                //  then write both clicks.

	    if(me.isAltDown())      {

	      secondClick =  me;
	      rfb.writePointerEvent(firstClick,middleclick);
	      rfb.writePointerEvent(secondClick,true);
	    }

                                //  if looking for left click and previous meta is not set
                                //  then it means that it is rightclick. Send two right clicks.


	    else if(me.isMetaDown() && !previousMetaPressed) {

	      secondClick =  me;
	      rfb.writePointerEvent(firstClick,middleclick);
	      rfb.writePointerEvent(secondClick,middleclick );
	      processingPressMouse = false;
	    } else {

	      secondClick =  me;

	      //condition for a middle click
                                   
	      if((test.abs(firstClick.getX()-secondClick.getX())) < 3  
		 && (test.abs(firstClick.getY()-secondClick.getY())) < 3) {

		middleclick = true;                                      
		rfb.writePointerEvent(firstClick,middleclick );
		processingPressMouse = false;
	      }

	      else {
		rfb.writePointerEvent(firstClick,middleclick );
		rfb.writePointerEvent(secondClick,middleclick );
		processingPressMouse = false;
	      }
	    }


                                // change the state and stop the thread beacuse decison
                                // has been reached.
	    processstate=listenstate;
	    theTimeOut.stop();
	    break;


	  case checkrightclick :

                                //  if looking for right click and middle button pressed
                                //      then write both clicks.

	    if(me.isAltDown())  {

	      secondClick =  me;
	      rfb.writePointerEvent(firstClick,middleclick);
	      rfb.writePointerEvent(secondClick,true);

	      //  if looking for right click and previous meta is not set
	      //  then it means that it is rightclick.Test for middle button.
	    } else if(me.isMetaDown() &&!previousMetaPressed) {

	      secondClick =  me;
                                   
	      if((test.abs(firstClick.getX()-secondClick.getX())) < 3
		 &&(test.abs(firstClick.getY()-secondClick.getY())) < 3) {

		middleclick = true;
		rfb.writePointerEvent(firstClick,middleclick);
		processingPressMouse = false;
	      } else {
		rfb.writePointerEvent(firstClick,middleclick );
		rfb.writePointerEvent(secondClick,middleclick );
		processingPressMouse = false;
	      }
	    } else {

	      secondClick =  me;
	      rfb.writePointerEvent(firstClick,middleclick);
	      rfb.writePointerEvent(secondClick,middleclick);
	      processingPressMouse = false;
	    }

                                // change the state and stop the thread beacuse decison
                                // has been reached.

	    processstate=listenstate;
	    theTimeOut.stop();
	    break;
	  }

	  break;

	case MouseEvent.MOUSE_MOVED:
	case MouseEvent.MOUSE_DRAGGED:

	  if(processingPressMouse)
	    Thread.sleep(60);

	  if (v.gotFocus) {
	    requestFocus();
	  }

	  rfb.writePointerEvent(e,false);
	  break;

	case MouseEvent.MOUSE_RELEASED:
	  //    Actually this was a big problem that a decison is reached
	  //    about a click in 30 msec while there respective released events
	  //    are detected earlier that is why here we make the thread to sleep
	  //    for 60 msec.

	  // in case of a middle click we have to ignore the next click.

	  MouseEvent me1 = (MouseEvent) e;

	  if(ignorenextclick) {
	    ignorenextclick = false;
	    break;
	  }

	  if(middleclick) {
	    ignorenextclick = true;
	    rfb.writePointerEvent(e,middleclick);
	    middleclick = false;
	    break;
	  }

	  if(processingPressMouse)        //sleep this thread for 60 msec.
	    Thread.sleep(60);

	  if (v.gotFocus) {
	    requestFocus();
	  }

	  if(me1.isMetaDown()) {
	    previousMetaPressed = false;
	  }

	  rfb.writePointerEvent(e,middleclick);
	  break;

	case KeyEvent.KEY_PRESSED:
	case KeyEvent.KEY_RELEASED:

	  rfb.writeKeyEvent(e);
	  break;

	  // Following is an event that has been defined for timeout. If
	  // within 50 msec 2 click does not happen then determine in
	  // which state we are and then write those events.


	case AWTEvent.RESERVED_ID_MAX+2:
	  switch(processstate) {

	  case listenstate :

	    processingPressMouse = false;
	    previousMetaPressed = false;
	    middleclick = false;
	    ignorenextclick = false;
                                //      up events.

	    theTimeOut.stop();
	    break;

	  case checkleftclick :
	    rfb.writePointerEvent(firstClick,middleclick);
	    processstate = listenstate;
	    processingPressMouse = false;
	    previousMetaPressed = false;
	    middleclick = false;
	    ignorenextclick = false;

	    theTimeOut.stop();
	    break;

	  case checkrightclick :
	    rfb.writePointerEvent(firstClick,middleclick);
	    processstate = listenstate;
	    processingPressMouse = false;
	    previousMetaPressed = false;
	    middleclick = false;
	    ignorenextclick = false;
	    theTimeOut.stop();
	    break;
	  }

	  break;
	} // end switch
      } catch (Exception f) {
	f.printStackTrace();
      }
    }
  }
}
