/***********************************************
     Author :      Muddassar Farooq
     Organizaton:  The University of New South Wales

     Date Started : October 1, 1998
     Date Ended   : November 22 , 1998.
     Main Module : vncviewer.java

     other Modules : animatedMemoryImageSource.java,
     authenticationPanel.java,ButtonPanel.java,
     clipBoardFrame.java, DesCipher.java,optionsFrame.java,
     rfbProto.java, TimeEvent.java, TimeOut.java, vncCanvas.java,

     This programs has been modified under GNU license as allowed by                                      
     its distributor Olivetti & Oracle Research Laboratory. The purpose
     was to upgrade this software to new java 1.1 event model and get
     rid of all deprecated APIs. Moreover to make this software more an
     application than an applet. Following modifications are done in the applet
     
     1. One can start it as an application by typing
        java vncviewer host:display.
     2. One can send a middle click as a double button click
     3. Scroll Pane is added to take care of large display area.
     4. Software is documented in a manner that conforms to principle
        of software Engineering so that its understanding and modifications
        become easy.


****************************************************************************/


//
//  Copyright (C) 1997, 1998 Olivetti & Oracle Research Laboratory
//
//  This is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This software is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this software; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
//  USA.
//

//
// vncviewer.java - the VNC viewer applet.  This class mainly just sets up the
// user interface, leaving it to the vncCanvas to do the actual rendering of
// a VNC desktop.
//

import java.awt.*;
import java.io.*;
import java.awt.event.*;
import java.lang.String;

public class vncviewer extends java.applet.Applet implements java.lang.Runnable
{
    boolean inAnApplet = true;

    Frame f;
    String[] mainArgs;
    String host;

    int port;
    rfbProto rfb;

    Thread rfbThread;
    GridBagLayout gridbag;
    GridBagConstraints gbc;
    ScrollPane pane;

    ButtonPannel buttonpanel;
    optionsFrame options;
    clipboardFrame clipboard;
    authenticationPanel authenticator;

    //
    // main() is called when run as a java program from the command line.  It
    // simply creates a frame and runs the applet inside it.
    //

    public static void main(String[] argv) {
        vncviewer v = new vncviewer();
        v.mainArgs = argv;
        v.inAnApplet = false;


        v.init();
        v.start();
    }

    //  The purpose of this contructor is to construct a frame and then
    //  close the window once user press at closing window button.
    //  In new event model, two events of lost focus and got focus are
    //  enabled so that we get rid of deprecated API'S.


    public vncviewer() {

        f = new Frame("VncViewer");	
        f.add("Center",this);
        

        f.addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) {
                System.exit(1);
            }
           }
        );
        this.enableEvents(FocusEvent.FOCUS_LOST | FocusEvent.FOCUS_GAINED);

    }


    //
    // init()
    //

    public void init() {

        //  If vncviewer is running as an application then we need to specify the
        //  argument as host:display so that the interface to user remains same.

        if (inAnApplet) {

            readParameters();
        }
        else {
            appreadparameter();
        }


        //  In layout of this applet GridBagLayOut has been used and its constraints
        //  are specified.

        gridbag = new GridBagLayout();
        gbc = new GridBagConstraints();

        options = new optionsFrame(this);
        clipboard = new clipboardFrame(this);
        authenticator = new authenticationPanel();
        buttonpanel = new ButtonPannel(this);

        rfbThread = new Thread(this);
        rfbThread.start();
    }

    public void update(Graphics g) {
    }

    //
    // run() - executed by the rfbThread to deal with the RFB socket.
    //

    public void run() {

        // Setting the layout for applet as gridbag.

        if(!inAnApplet)
          f.setLayout(gridbag);
        else
          setLayout(gridbag);
       
 
        gbc.gridwidth = GridBagConstraints.REMAINDER;
        gbc.anchor = GridBagConstraints.NORTHWEST;
        gridbag.setConstraints(buttonpanel, gbc);
       
        if(!inAnApplet) 
           f.add(buttonpanel);
        else
           add(buttonpanel);

        try {
            connectAndAuthenticate();

            doProtocolInitialisation();

            // These lines actually are there to add a scrollbar pane in the canvas
            // in case they are needed. In fact in older version if one tries to
            // run java vncviewer as an application then there was no nmechanism
            // of showing large screens on small one.


            //  The approach taken is very simple i.e. one gets the screen size and then
            //  determines the size of rfb server. The scrollPane size is set to smallest
            //  of both.

            if (!inAnApplet) {

                Dimension sizeScreen =  Toolkit.getDefaultToolkit().getScreenSize();

                int scrollWidth = sizeScreen.width;
                int scrollHeight = sizeScreen.height;


                if (rfb.framebufferWidth < scrollWidth)
                    scrollWidth = rfb.framebufferWidth;

                if (rfb.framebufferHeight < scrollHeight)
                    scrollHeight = rfb.framebufferHeight;

                pane = new ScrollPane(ScrollPane.SCROLLBARS_AS_NEEDED);
                pane.setSize(scrollWidth, scrollHeight - 70);
                gbc.weightx = 1.0;
                gbc.weighty = 1.0;
                  
                gridbag.setConstraints(pane, gbc);
                f.add(pane);
            }


            vncCanvas vc = new vncCanvas(this);
            if(inAnApplet) {
                gbc.weightx = 1.0;
                gbc.weighty = 1.0;
                gridbag.setConstraints(vc, gbc);
            }


            //  If vncviewer is running as an application then canvas needs to be added
            //  in Scrollable pane else simply add to the applet

            if (!inAnApplet) {
                pane.add(vc);
            }
            else {
                add(vc);
            }


            //  If vncviewer is running as an application then display the name of
            //  the desktop as title of frame.
            //


            if (!inAnApplet) {
                f.setTitle(rfb.desktopName);
                f.pack();
            } else {
                validate();
            }

            //  This has been common obervation that if user is given a chance to modify
            //  option especially after getting connected to remote terminal then in
            //  cases of coding that may lead to complete crash of a system. The reason
            //  of that was since double click was not implemented as a middle one, user
            //  was given control of changing mouse buttons. Since in new version double
            //  click has been implemented, hence optons button is disabled for smooth
            //  running of system.

            buttonpanel.optionsButton.setEnabled(false);
            buttonpanel.disconnectButton.setEnabled(true);
            buttonpanel.clipboardButton.setEnabled(true);
            buttonpanel.ctrlAltDelButton.setEnabled(true);

            //  Once all initialization and authentication has been taken place then
            //  start processing the information in a normal way.


            vc.processNormalProtocol();

        }
        catch (Exception e) {
            e.printStackTrace();
            fatalError(e.toString());
        }

    }


    //
    // Connect to the RFB server and authenticate the user.
    //

    void connectAndAuthenticate() throws IOException {

        GridBagConstraints gbc = new GridBagConstraints();
        gbc.gridwidth = GridBagConstraints.REMAINDER;
        gbc.anchor = GridBagConstraints.NORTHWEST;
        gbc.weightx = 1.0;
        gbc.weighty = 1.0;
        gbc.ipadx = 100;
        gbc.ipady = 50;
        gridbag.setConstraints(authenticator, gbc);
        if(!inAnApplet) 
           f.add(authenticator);
        else
           add(authenticator);
 
        validate();
        if (!inAnApplet) {
            f.pack();
            f.show();
        }


        boolean authenticationDone = false;

        //      In this situation till the time authentication has not been verified,
        //      iteration continues. However there is an important concept that
        //      control flow does not go ahead from here until user clicks ok button
        //      or presses enter in the textfield. At that time the this object of
        //      monitor is notified.
        //

        while (!authenticationDone) {

            synchronized (authenticator) {
                try {
                    authenticator.wait();
                } catch (InterruptedException e) {
                }
            }

            rfb = new rfbProto(host, port, this);

            rfb.readVersionMsg();

            System.out.println( "RFB server supports protocol version " +  rfb.serverMajor + "." + rfb.serverMinor);

            rfb.writeVersionMsg();

            switch (rfb.readAuthScheme()) {

                case rfb.NoAuth:
                    System.out.println("No authentication needed");
                    authenticationDone = true;
                    break;

                case rfb.VncAuth:

                    //In case of VncAuth the authentication scheme used is simple
                    //challenge/response scheme, i.e. server sends the challenge which client
                    //encrypts using a shared key and then waits for response from the server.

                    byte[] challenge = new byte[16];
                    rfb.is.readFully(challenge);

                    //In case password is more than 8 characyers then truncate it to 8.


                    String pw = authenticator.password.getText();
                    int length = pw.length();
                    if (length > 8)
                        pw = pw.substring(0, 8); // truncate to 8 chars

                    if (length == 0) {
                        authenticator.retry();
                        break;
                    }

                    // In order to get rid of deprecated getBytes, a simple thing is done as the
                    // new getBytes throws exception if byte[] array is smaller or larger than
                    // String size. If string is smaller then we simply define two arrays so
                    // that this exception can be avoided.

                    byte[] key = new byte[8];
                    byte[] tempkey = new byte[length];

                    if (length < 8) {
                        tempkey = pw.getBytes();

                        for (int i = 0; i < length; i++) {
                            key[i] = tempkey[i];
                        }

                        for (int i = length; i < 8; i++) {
                            key[i] = (byte) 0;
                        }
                    }
                    else {
                        key = pw.getBytes();
                    }


                    //Call decipher class in order to encrypt the challenge.

                    DesCipher des = new DesCipher(key);

                    des.encrypt(challenge, 0, challenge, 0);
                    des.encrypt(challenge, 8, challenge, 8);

                    rfb.os.write(challenge);

                    int authResult = rfb.is.readInt();

                    //Result of encrypting with shared key arrived.

                    switch (authResult) {
                        case rfb.VncAuthOK:

                            System.out.println("VNC authentication succeeded");
                            authenticationDone = true;
                            break;

                        case rfb.VncAuthFailed:
                            System.out.println("VNC authentication failed");
                            authenticator.retry();
                            break;

                        case rfb.VncAuthTooMany:
                            throw new IOException(
                                    "VNC authentication failed - " + "too many tries");

                        default:
                            throw new IOException( "Unknown VNC authentication result " +  authResult);
                    }
                    break;
            }
        }

      if(!inAnApplet) 
         f.remove(authenticator);
      else
        remove(authenticator); 
    }


    //
    // Do the rest of the protocol initialisation.
    //

    void doProtocolInitialisation() throws IOException {
        System.out.println("sending client init");

        rfb.writeClientInit();

        rfb.readServerInit();

        System.out.println("Desktop name is " + rfb.desktopName);
        System.out.println("Desktop size is " + rfb.framebufferWidth + " x " + rfb.framebufferHeight);

        setEncodings();
    }


    //
    // setEncodings() - send the current encodings from the options frame
    // to the RFB server.
    //

    void setEncodings() {
        try {
            if ((rfb != null) && rfb.inNormalProtocol) {
                rfb.writeSetEncodings(options.encodings,
                        options.nEncodings);
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
    }


    //
    // setCutText() - send the given cut text to the RFB server.
    //

    void setCutText(String text) {
        try {
            if ((rfb != null) && rfb.inNormalProtocol) {
                rfb.writeClientCutText(text);
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
    }


    //
    // Respond to an action i.e. button press
    //



    //
    // Detect when the focus goes in and out of the applet.  See
    // vncCanvas.handleEvent() for details of why this is necessary.
    //

    boolean gotFocus = false;

    //  processFocusEvent(FocusEvent fe). This is new 1.1 Model in which
    //  two events for losing and gaining focus are processed.
    //

    public void processFocusEvent(FocusEvent fe) {

        int typeEvent = fe.getID();

        if (typeEvent == FocusEvent.FOCUS_LOST)
            gotFocus = false;

        else if (typeEvent == FocusEvent.FOCUS_GAINED) {
            gotFocus = true;
        }
    }





    //
    // encryptBytes() - encrypt some bytes in memory using a password.  Note that
    // the mapping from password to key must be the same as that used on the rfb
    // server side.   

    //
    // Note also that IDEA encrypts data in 8-byte blocks, so here we will ignore
    // any data beyond the last 8-byte boundary leaving it to the calling
    // function to pad the data appropriately.
    //

    void encryptBytes(byte[] bytes, String passwd) {
        byte[] key = new byte[8];

        key = passwd.getBytes();

        for (int i = passwd.length(); i < 8; i++) {
            key[i] = (byte) 0;
        }

        DesCipher des = new DesCipher(key);

        des.encrypt(bytes, 0, bytes, 0);
        des.encrypt(bytes, 8, bytes, 8);
    }


    //
    //    appreadparameter(). This is a procedure that is used to read
    //    host:display parameter. Simply it does a substring and then
    //    used standard string fucntions to read that.

    public void appreadparameter() {
        try {
            String mainArgsvalue = mainArgs[0];
            int index = mainArgsvalue.indexOf(':');

            // In case no argument has been specified.

            if (index == 0)
                fatalError("Specify Argument as Host:Display");

            host = mainArgsvalue.substring(0, index);
            port = 5900 +
                    Integer.parseInt(mainArgsvalue.substring(index + 1));
        } catch (IndexOutOfBoundsException strEx) {
            fatalError("Specify Argument as Host:Display");
        }
        catch (NumberFormatException NumEx) {
            fatalError("Specify Argument as Host:Display");
        }
    }


    //
    // readParameters() - read parameters from the html source or from the
    // command line.  On the command line, the arguments are just a sequence of
    // param_name/param_value pairs where the names and values correspond to
    // those expected in the html applet tag source.
    //

    public void readParameters() {
        host = readParameter("HOST", !inAnApplet);
        if (host == null) {
            host = getCodeBase().getHost();
            if (host.equals("")) {
                fatalError("HOST parameter not specified");
            }
        }

        String s = readParameter("PORT", true);
        port = Integer.parseInt(s);
    }


    // readParameter((String name, boolean required): This method actually
    // simply reads command line parameters and if that parameter is required
    // and not specified then it throws fatal error.

    public String readParameter(String name, boolean required) {
        if (inAnApplet) {
            String s = getParameter(name);
            if ((s == null) && required) {
                fatalError(name + " parameter not specified");
            }
            return s; 
        }

        for (int i = 0; i < mainArgs.length; i += 2) {
            if (mainArgs[i].equalsIgnoreCase(name)) {
                try {
                    return mainArgs[i + 1];
                } catch (Exception e) {
                    if (required) {
                        fatalError(name + " parameter not specified");
                    }
                    return null;
                }
            }
        }

        if (required) {
            fatalError(name + " parameter not specified");
        }
        return null;
    }

    //
    // fatalError() - print out a fatal error message and if running as an
    // applet then print disconnected and stop the thread.
    //

    public void fatalError(String s) {
        System.out.println(s);

        if (inAnApplet) {
            removeAll();
            Label l = new Label(s);

            setLayout(new FlowLayout(FlowLayout.LEFT, 30, 30));
            add(l);
            validate();
            Thread.currentThread().stop();
        } else {
            System.exit(1);
        }
    }
}






